/*-
 * Copyright (c) 2002 Olivier Houchard
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer 
 *    in this position and unchanged.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $FreeBSD: src/sys/compat/linux/linux_file.c,v 1.64 2002/06/02 20:05:41 schweikh Exp $
 */

#include "opt_compat.h"

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/conf.h>
#include <sys/dirent.h>
#include <sys/fcntl.h>
#include <sys/file.h>
#include <sys/filedesc.h>
#include <sys/lock.h>
#include <sys/malloc.h>
#include <sys/mount.h>
#include <sys/mutex.h>
#include <sys/proc.h>
#include <sys/sysproto.h>
#include <sys/tty.h>
#include <sys/vnode.h>
#include <sys/namei.h>
#include <sys/stat.h>
#include <sys/syscallsubr.h>
#include <ufs/ufs/extattr.h>
#include <ufs/ufs/quota.h>
#include <ufs/ufs/ufsmount.h>

#include <compat/netbsd/netbsd_proto.h>
#include <compat/netbsd/netbsd_util.h>

static int setfflags(struct thread *td, struct vnode *, int);

static int
setfflags(td, vp, flags)
	struct thread *td;
	struct vnode *vp;
	int flags;
{
	int error;
	struct mount *mp;
	struct vattr vattr;

	/*
	 * Prevent non-root users from setting flags on devices.  When
	 * a device is reused, users can retain ownership of the device
	 * if they are allowed to set flags and programs assume that
	 * chown can't fail when done as root.
	 */
	if (vp->v_type == VCHR || vp->v_type == VBLK) {
		error = suser_cred(td->td_ucred, PRISON_ROOT);
		if (error)
			return (error);
	}

	if ((error = vn_start_write(vp, &mp, V_WAIT | PCATCH)) != 0)
		return (error);
	VOP_LEASE(vp, td, td->td_ucred, LEASE_WRITE);
	vn_lock(vp, LK_EXCLUSIVE | LK_RETRY, td);
#ifdef MAC
	error = mac_check_vnode_setflags(td->td_ucred, vp, vattr.va_flags);
	if (error == 0) {
#endif
		VATTR_NULL(&vattr);
		vattr.va_flags = flags;
		error = VOP_SETATTR(vp, &vattr, td->td_ucred, td);
#ifdef MAC
	}
#endif
	VOP_UNLOCK(vp, 0, td);
	vn_finished_write(mp);
	return (error);
}


#ifdef COMPAT_43
int
onetbsd_creat(struct thread *td, struct onetbsd_creat_args *args)
{
    char *path;
    int	error;

    LCONVPATHCREAT(td, args->path, &path);
    error = kern_open(td, path, UIO_SYSSPACE, O_WRONLY | O_CREAT | O_TRUNC,
		    args->mode);
    LFREEPATH(path);
    return (error);
}
#endif /*!COMPAT_43*/

int
netbsd_chflags(struct thread *td, struct netbsd_chflags_args *args)
{
    char		*path;
    int			error;
    struct nameidata 	nd;

    LCONVPATHEXIST(td, args->path, &path);
    NDINIT(&nd, LOOKUP, FOLLOW, UIO_SYSSPACE, path, td);
    if ((error = namei(&nd)) != 0) 
	   goto done; 
    NDFREE(&nd, NDF_ONLY_PNBUF);
    error = setfflags(td, nd.ni_vp, args->flags);
    vrele(nd.ni_vp);
done:
    LFREEPATH(path);
    return (error);
}

int
netbsd_nstat(struct thread *td, struct netbsd_nstat_args *args)
{
    struct stat		sb;
    struct nstat	nsb;
    struct nameidata	nd;
    char		*path;
    int			error;

    LCONVPATHEXIST(td, args->path, &path);
    NDINIT(&nd, LOOKUP,FOLLOW | LOCKLEAF | NOOBJ, UIO_SYSSPACE, path, td);
    if ((error = namei(&nd)))
	    goto done;
    NDFREE(&nd, NDF_ONLY_PNBUF);
    error = vn_stat(nd.ni_vp, &sb, td->td_ucred, NOCRED, td);
    vput(nd.ni_vp);
    if (error)
	    goto done;
    cvtnstat(&sb, &nsb);
    error = copyout(&nsb, args->ub, sizeof(nsb));
done:
    LFREEPATH(path);
    return (error);
}

int
netbsd_execve(struct thread *td, struct netbsd_execve_args *args)
{
    struct execve_args fbsd_args;
    caddr_t	sg;
    int		error;
    
    sg = nbsd_stackgap_init();
    mtx_lock(&Giant);
    CHECKALTEXIST(td, &sg, args->fname);
    mtx_unlock(&Giant);
    fbsd_args.fname = args->fname;
    fbsd_args.argv = args->argv;
    fbsd_args.envv = args->envv;
    error = execve(td, &fbsd_args);
    return (error);
}

int
netbsd_nlstat(struct thread *td, struct netbsd_nlstat_args *args)
{
    struct vnode	*vp;
    struct stat		sb;
    struct nstat	nsb;
    struct nameidata	nd;
    char		*path;
    int			error;    

    LCONVPATHEXIST(td, args->path, &path);
    NDINIT(&nd, LOOKUP, NOFOLLOW | LOCKLEAF | NOOBJ, UIO_SYSSPACE, path, td);
    if ((error = namei(&nd)) != 0)
	    goto done;
    vp = nd.ni_vp;
    NDFREE(&nd, NDF_ONLY_PNBUF);
    error = vn_stat(vp, &sb, td->td_ucred, NOCRED, td);
    vput(vp);
    if (error)
	    goto done;
    cvtnstat(&sb, &nsb);
    error = copyout(&nsb, args->ub, sizeof(nsb));
done:
    LFREEPATH(path);
    return (error);
}


int
netbsd_open(struct thread *td, struct netbsd_open_args *args)
{
    int 	error;
    char	*path;

    if (args->flags & O_CREAT)
	LCONVPATHCREAT(td, args->path, &path);
    else
	LCONVPATHEXIST(td, args->path, &path);
    error = kern_open(td, path, UIO_SYSSPACE, args->flags, args->mode);
    LFREEPATH(path);
    return (error);
}

int
netbsd_access(struct thread *td, struct netbsd_access_args *args)
{
	char	*path;
	int	error;

	LCONVPATHEXIST(td, args->path, &path);
	error = kern_access(td, path, UIO_SYSSPACE, args->flags);
	LFREEPATH(path);
	return (error);
}

int
netbsd_unlink(struct thread *td, struct netbsd_unlink_args *args)
{
	char	*path;
	int	error;

	LCONVPATHEXIST(td, args->path, &path);
	error = kern_unlink(td, path, UIO_SYSSPACE);
	LFREEPATH(path);
	return (error);	
}

int
netbsd_chdir(struct thread *td, struct netbsd_chdir_args *args)
{
	char	*path;
	int	error;

	LCONVPATHEXIST(td, args->path, &path);
	error = kern_chdir(td, path, UIO_SYSSPACE);
	LFREEPATH(path);
	return (error);
}

int
netbsd_chmod(struct thread *td, struct netbsd_chmod_args *args)
{
	char	*path;
	int	error;

	LCONVPATHEXIST(td, args->path, &path);
	error = kern_chmod(td, path, UIO_SYSSPACE, args->mode);
	LFREEPATH(path);
	return (error);
}

int
netbsd_mkdir(struct thread *td, struct netbsd_mkdir_args *args)
{
	char	*path;
	int	error;

	LCONVPATHCREAT(td, args->path, &path);
	error = kern_mkdir(td, path, UIO_SYSSPACE, args->mode);
	LFREEPATH(path);
	return (error);
}

int
netbsd_rmdir(struct thread *td, struct netbsd_rmdir_args *args)
{
	char	*path;
	int	error;

	LCONVPATHEXIST(td, args->path, &path);
	error = kern_rmdir(td, path, UIO_SYSSPACE);
	LFREEPATH(path);
	return (error);
}

int
netbsd_rename(struct thread *td, struct netbsd_rename_args *args)
{
	char	*pathfrom, *pathto;
	int	error;

	LCONVPATHEXIST(td, args->from, &pathfrom);
	error = netbsd_emul_convpath(td, args->to, UIO_USERSPACE, &pathto, 1);
	if (!pathto) {
		LFREEPATH(pathfrom);
		return (error);
	}
	error = kern_rename(td, pathfrom, pathto, UIO_SYSSPACE);
	LFREEPATH(pathfrom);
	LFREEPATH(pathto);
	return (error);
}

int
netbsd_symlink(struct thread *td, struct netbsd_symlink_args *args)
{
	char	*path, *plink;
	int	error;

	LCONVPATHEXIST(td, args->path, &path);
	error = netbsd_emul_convpath(td, args->link, UIO_USERSPACE, &plink ,1);
	if (plink == NULL) {
		LFREEPATH(path);
		return (error);
	}

	error = kern_symlink(td, path, plink, UIO_SYSSPACE);
	LFREEPATH(path);
	LFREEPATH(plink);
	return (error);
}

int
netbsd_readlink(struct thread *td, struct netbsd_readlink_args *args)
{
	char	*path;
	int	error;

	LCONVPATHEXIST(td, args->path, &path);

	error = kern_readlink(td, path, UIO_SYSSPACE, args->buf, UIO_USERSPACE,
			args->count);
	LFREEPATH(path);
	return (error);
}

int
netbsd_truncate(struct thread *td, struct netbsd_truncate_args *args)
{
	char	*path;
	int	error;

	LCONVPATHEXIST(td, args->path, &path);
	error = kern_truncate(td, path, UIO_SYSSPACE, args->length);
	LFREEPATH(path);
	return (error);
}

int
netbsd_link(struct thread *td, struct netbsd_link_args *args)
{
    char	*path, *plink;
    int		error;

    LCONVPATHEXIST(td, args->path, &path);
    error = netbsd_emul_convpath(td, args->link, UIO_USERSPACE, &plink, 1);
    if (plink == NULL) {
	    LFREEPATH(path);
	    return (error);
    }
    error = kern_link(td, path, plink, UIO_SYSSPACE);
    LFREEPATH(path);
    LFREEPATH(plink);
    return (error);
}

int
netbsd_chown(struct thread *td, struct netbsd_chown_args *args)
{
	char	*path;
	int	error;

	LCONVPATHEXIST(td, args->path, &path);

	error = kern_chown(td, path, UIO_SYSSPACE, args->uid, args->gid);
	LFREEPATH(path);
	return (error);
}

int
netbsd_lchown(struct thread *td, struct netbsd_lchown_args *args)
{
	char	*path;
	int	error;

	LCONVPATHEXIST(td, args->path, &path);

	error = kern_lchown(td, path, UIO_SYSSPACE, args->uid, args->gid);
	LFREEPATH(path);
	return (error);
}
